#!/bin/bash
#
# Simple ProFTPd account and ticketing manager toolkit.
# Part of proftpm package Sorry about my terrible English! :-)

# Including headers
EXIT_ERR=1
HEADERS="/etc/proftpm/proftpm.conf"
MSG_PRM_NOHEADERS="Header file $HEADERS is not found - stopped."
MSG_PRM_NOLANGUAGES="Language file $TEMPLATELANG is not found - stopped."
if [ ! -r "$HEADERS" ]; then echo -e "$MSG_PRM_NOHEADERS" >&2; exit $EXIT_ERR
    else . $HEADERS
fi
# Including defaults (if any)
if [ -r "$DEFAULTS" ]; then . $DEFAULTS; fi
# Including language file
if [ -r "$LANGUAGEDIR/$DEFAULT_LANGUAGE/messages" ]; then
    . "$LANGUAGEDIR/$DEFAULT_LANGUAGE/messages"
else
    if [ ! -r "$TEMPLATELANG" ]; then echo -e "$MSG_PRM_NOLANGUAGES" >&2; exit $EXIT_ERR
        else . "$TEMPLATELANG"
    fi
fi

#
# Functions
#

# proftpm_createaccount
# Creates a permanent FTP account as root or member of adminusers or admingroup.
function proftpm_createaccount {

    # permissions check: admin and mysql are required
    isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi

    # username is mandatory, formally checked and must be unused
    local username=$1; shift
    if [ -z "$username" ]; then log "$MSG_PRM_CREATEACC_MISSINGUSER"; return $EXIT_ERR; fi
    if [ ! -z "`echo "$username" | $SED "s/$MASKUSER//"`" ]; then log "$MSG_PRM_CREATEACC_WRONGUSER $username"; return $EXIT_ERR; fi
    if [ ! -z "`proftpm_listaccounts $username`" ]; then log "$MSG_PRM_CREATEACC_EXISTINGUSER" "$username"; return $EXIT_ERR; fi
    # email is mandatory and formally checked
    local email="$1"; shift
    if [ -z "$email" ]; then log "$MSG_PRM_CREATEACC_MISSINGEMAIL"; return $EXIT_ERR; fi
    if [ ! -z "`echo "$email" | $SED "s/$MASKEMAIL//"`" ]; then log "$MSG_PRM_CREATEACC_WRONGEMAIL $email"; return $EXIT_ERR; fi
    # password is optional but formally checked; generated when isn't given then encoded
    local password=$1; shift
    if [ -z "$password" ]; then 
        password="`$MAKEPASSWD`"
    else
        if [ ! -z "`echo "$password" | $SED "s/$MASKPASS//"`" ]; then log "$MSG_PRM_CREATEACC_WRONGPASS"; return $EXIT_ERR; fi
    fi
    local passhash="{md5}$(echo -n $password | $OPENSSL dgst -binary -md5 | $OPENSSL enc -base64)"
    local passclear="$(echo -n $password | $OPENSSL enc -base64 | $OPENSSL enc -base64)"
    # FTP root check 
    if [ ! -d "$FTPROOT" ]; then log "$MSG_PRM_CREATEACC_WRONGFTPROOT $FTPROOT"; return $EXIT_ERR; fi
    local homedir="$FTPROOT/$username"
    # Here we have valid user/email/pass/hash/homedir parameters.

    local RESULT_CODE=0
    # Account creation in MySQL table.
    local MYSQL_ADDUSER="INSERT INTO users SET userid='$username', email='$email', passwd='$passhash', clear='$passclear', uid=$FTPUID, gid=$FTPGID, homedir='$homedir';"
    log "$username" "$(echo "$MYSQL_ADDUSER" | $MYSQL 2>&1; RESULT_CODE=$?)" "$RESULT_CODE"
    # Mail notification on success (will send credentials in plain text!).
    if [ $RESULT_CODE -eq 0 ]; then
        export FTPURL password REPLYTO="$SYSADMIN" SYSADMIN username
        cat "$MAIL_PRM_CREATEDACC" | $ENVSUBST | $MAIL -a "$MAIL_PRM_CREATEDACC_MIME" -s "$MAIL_PRM_CREATEDACC_SUBJ" "$email"; RESULT_CODE=$?
        log "$MSG_PRM_MAILSENT $email $RESULT_CODE"
    fi

    return $RESULT_CODE
}

# proftpm_createticket ownername email [guestname [password [expiry]]]
# Creates a transient FTP account (ticket).
function proftpm_createticket {

    local RESULT_CODE=0

    # Permissions check: mysql required
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi

    # Existing ownername is mandatory.
    local parentid="$1"; shift
    if [ -z "$parentid" ]; then log "$MSG_PRM_CREATETICKET_MISSINGUSER"; return $EXIT_ERR; fi
    # Loads parent details into a simulated hash.
    # Some IFS tricks are used to avoid problems from fields containing spaces.
    local MYSQL_LISTACCOUNT="SELECT * FROM users WHERE userid='$parentid';"
    local ifs=$IFS; IFS=$MYSQL_IFS; local details=($(IFS=$ifs; echo "$MYSQL_LISTACCOUNT" | $MYSQL 2>/dev/null)); IFS=$ifs
    if [ -z "${details[0]}" ]; then log "$MSG_PRM_CREATETICKET_NOSUCHUSER $parentid"; return $EXIT_ERR; fi
    # A guest account cannot be parent.
    if [ ! -z "$(getdetail 'parentid')" -a "$(getdetail 'parentid')" != "NULL" ]; then
        log "$MSG_PRM_CREATETICKET_GUESTPARENT" $parentid; return $EXIT_ERR
    fi
    # Email is mandatory and formally checked
    local email="$1"; shift
    if [ -z "$email" ]; then log "$MSG_PRM_CREATETICKET_MISSINGEMAIL"; return $EXIT_ERR; fi
    if [ ! -z "`echo "$email" | $SED "s/$MASKEMAIL//"`" ]; then log "$MSG_PRM_CREATETICKET_WRONGEMAIL $email"; return $EXIT_ERR; fi
    # Guestname is optional but formally checked.
    # Admin rights required to specify guestname.
    local username="$1"; shift
    if [ ! -z "$username" ]; then
        isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_CREATETICKET_ADMINNEED"; return $EXIT_ERR; fi
        # Guest name formally checked and must be unused.
        if [ ! -z "`echo "$username" | $SED "s/$MASKUSER//"`" ]; then log "$MSG_PRM_CREATETICKET_WRONGUSER $username"; return $EXIT_ERR; fi
        if [ ! -z "`proftpm_listaccounts $username`" ]; then log "$MSG_PRM_CREATETICKET_EXISTINGUSER" "$username"; return $EXIT_ERR; fi
    else
        # Creates a pseudorandom guestname if is not given.
        username="$parentid"_"$(pseudorandom 6)"
        # Formally checked and must be unused.
        # On error aborted vith an obscure error message...
        if [ ! -z "`echo "$username" | $SED "s/$MASKUSER//"`" ]; then log "$MSG_PRM_CREATETICKET_OBSCURE"; return $EXIT_ERR; fi
        if [ ! -z "`proftpm_listaccounts $username`" ]; then log "$MSG_PRM_CREATETICKET_OBSCURE"; return $EXIT_ERR; fi
    fi
    # Password is optional but formally checked;
    # generated when isn't given then encoded.
    # Admin rights required to specify password.
    local password=$1; shift
    if [ ! -z "$password" ]; then 
        isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_CREATETICKET_ADMINNEED"; return $EXIT_ERR; fi
        # Format check need.
        if [ ! -z "`echo "$password" | $SED "s/$MASKPASS//"`" ]; then log "$MSG_PRM_CREATETICKET_WRONGPASS"; return $EXIT_ERR; fi
    else
        # Creates a random password if is not given.
        password="`$MAKEPASSWD`"
    fi
    local passhash="{md5}$(echo -n $password | $OPENSSL dgst -binary -md5 | $OPENSSL enc -base64)"
    local passclear="$(echo -n $password | $OPENSSL enc -base64 | $OPENSSL enc -base64)"
    # Expiry date is optional but validated.
    # Admin rights required to specify expiry.
    local expiry="$1"; shift
    if [ ! -z "$expiry" ]; then 
        isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_CREATETICKET_ADMINNEED"; return $EXIT_ERR; fi
        # Expiry validation via system 'date' command.
        if [ -z "$($DATE -d "$expiry" 2>/dev/null)" ]; then log "$MSG_PRM_CREATETICKET_WRONGEXPIRY $expiry"; return $EXIT_ERR; fi
    else
        # Sets default lifetime if is not given.
        expiry="$DEFAULT_EXPIRY"
    fi
    expiry="$($DATE '+%Y-%m-%d %H:%M:%S' -d "$expiry" 2>/dev/null)"
    # All details was processed.

    # We need an existing homedir for parent.within $FTPROOT
    local homedir="$(getdetail 'homedir')"
    if [ -z "$homedir" -o $((( ${#homedir}-${#FTPROOT}-1 ))) -le 0 -o $(expr match "$homedir" "$FTPROOT") -eq 0  ]; then
        log "$MSG_PRM_CREATETICKET_OBSCURE"; return $EXIT_ERR; fi
    # New unique FTP home within parents incoming area: timestamp_masked-email-address
    local guestfolder="$INCOMING$($DATE '+%Y%m%d%H%M%S')_FROM_$(echo -ne "$email" | $SED "s/@/_AT_/g" | $SED "s/\./_DOT_/g")"
    homedir="$homedir/$guestfolder"

    # Account creation in MySQL table.
    local MYSQL_ADDUSER="INSERT INTO users SET userid='$username', email='$email', passwd='$passhash', clear='$passclear', parentid='$parentid', expiry='$expiry', uid=$FTPUID, gid=$FTPGID, homedir='$homedir';"
    log "$username" "$(echo "$MYSQL_ADDUSER" | $MYSQL 2>&1; RESULT_CODE=$?)" "$RESULT_CODE"
    # Mail notifications on success.
    if [ $RESULT_CODE -eq 0 ]; then
        # For guest user (will send credentials in plain text!).
        local restriction="$MSG_PRM_CREATETICKET_RESTRICT $parentid"
        export expiry FTPURL password REPLYTO="$SYSADMIN" restriction SYSADMIN username
        cat "$MAIL_PRM_CREATEDTICKET_GUEST" | $ENVSUBST | $MAIL -a "$MAIL_PRM_CREATEDTICKET_GUEST_MIME" -s "$MAIL_PRM_CREATEDTICKET_GUEST_SUBJ" "$email"; RESULT_CODE=$?
        log "$MSG_PRM_MAILSENT $email $RESULT_CODE"
        # For host user (owner).
        local guestemail="$email"
        email="$(getdetail 'email')"
        export expiry FTPURL guestemail guestfolder REPLYTO="$SYSADMIN" parentid SYSADMIN username
        cat "$MAIL_PRM_CREATEDTICKET_HOST" | $ENVSUBST | $MAIL -a "$MAIL_PRM_CREATEDTICKET_HOST_MIME" -s "$MAIL_PRM_CREATEDTICKET_HOST_SUBJ" "$email"; RESULT_CODE=$?
        log "$MSG_PRM_MAILSENT $email $RESULT_CODE"
    fi

    return $RESULT_CODE
}

# proftpm_listaccounts
# With username and password given authenticates the user 
# (on success lists all details).
# With username given lists all details of the user.
# Without parameter lists all permanent accounts.
function proftpm_listaccounts {

    # Permissions check: mysql required.
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi
    # Non-admins allowed to see only public details.
    local details="*"
    isadmin; if [ $? -eq $EXIT_ERR ]; then details="$PUBLIC_DETAILS"; fi

    local username=$1; shift            # Optional user ID
    local password=$1; shift            # Optional password
    local passhash="{md5}$(echo -n $password | $OPENSSL dgst -binary -md5 | $OPENSSL enc -base64)"

    local RESULT_CODE=0

    if [ -z "$username" ]; then
        # No parameters: lists all accounts.
        local MYSQL_LISTACCOUNTS="SELECT $details FROM users ORDER BY userid;"
        echo "$MYSQL_LISTACCOUNTS" | $MYSQL 2>/dev/null; RESULT_CODE=$?
    else
        if [ -z "$password" ]; then
            # No password: lists the user if exists.
            local MYSQL_LISTACCOUNT="SELECT $details FROM users WHERE userid='$username';"
            echo "$MYSQL_LISTACCOUNT" | $MYSQL 2>/dev/null; RESULT_CODE=$?
        else
            # Username and password: authenticates it.
            local MYSQL_CHECKACCOUNT="SELECT $details FROM users WHERE userid='$username' AND passwd='$passhash';"
            echo "$MYSQL_CHECKACCOUNT" | $MYSQL 2>/dev/null; RESULT_CODE=$?
        fi
    fi
    return $RESULT_CODE
}

# proftpm_listtickets [owner]
# With owner ID given lists tickets corresponding to this owner.
# Without parameter lists all tickets.
function proftpm_listtickets {

    # Permissions check: mysql required
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi
    # Non-admins allowed to see only public details.
    local details="*"
    isadmin; if [ $? -eq $EXIT_ERR ]; then details="$PUBLIC_DETAILS"; fi

    local owner=$1; shift               # Optional parent ID
    local RESULT_CODE=0

    if [ -z "$owner" ]; then
        local MYSQL_LISTTICKETS="SELECT $details FROM users WHERE parentid IS NOT NULL AND NOT parentid='' ORDER BY userid;"
        echo "$MYSQL_LISTTICKETS" | $MYSQL 2>/dev/null; RESULT_CODE=$?
    else
        local MYSQL_LISTTICKETSOWNED="SELECT $details FROM users WHERE parentid='$owner' ORDER BY userid;"
        echo "$MYSQL_LISTTICKETSOWNED" | $MYSQL 2>/dev/null; RESULT_CODE=$?
    fi
    return $RESULT_CODE
}

# proftpm_modifyaccount username [old_password [email [password [expiry]]]]
# Modifies details of an existing account or sends a password reminder email.
function proftpm_modifyaccount {

    # Permissions check: mysql required
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi

    local RESULT_CODE=0

    # Existing username is mandatory.
    local username="$1"; shift
    if [ -z "$username" ]; then log "$MSG_PRM_MODIFYACC_MISSINGUSER"; return $EXIT_ERR; fi
    # Loads user details into a simulated hash.
    # Some IFS tricks are used to avoid problems from fields containing spaces.
    local MYSQL_LISTACCOUNT="SELECT * FROM users WHERE userid='$username';"
    local ifs=$IFS; IFS=$MYSQL_IFS; local details=($(IFS=$ifs; echo "$MYSQL_LISTACCOUNT" | $MYSQL 2>/dev/null)); IFS=$ifs
    if [ -z "${details[0]}" ]; then log "$MSG_PRM_MODIFYACC_NOSUCHUSER $username"; return $EXIT_ERR; fi

    # Password is optional.
    local password="$1"; shift
    isadmin; if [ $? -eq $EXIT_ERR ]; then
        # Nonadmin users may get a password reminder giving no (or empty) password.
        # OK, this is a nasty sechole.
        if [ -z "$password" ]; then
            password="$(getdetail 'clear' | $OPENSSL enc -d -base64 | $OPENSSL enc -d -base64)"
            # Gets other details.
            local email="$(getdetail 'email')"
            local expiry="$(getdetail 'expiry')"; if [ -z "$expiry" -o "$expiry" = "NULL" -o "$expiry" = "0000-00-00 00:00:00" ]; then expiry="$MSG_PRM_MODIFYACC_NOEXPIRY"; fi
            local parentid="$(getdetail 'parentid')"
            local restriction
            if [ -z "$parentid" -o "$parentid" = "NULL" ]; then
                restriction="$MSG_PRM_MODIFYACC_NORESTRICT"
            else
                restriction="$MSG_PRM_MODIFYACC_RESTRICT $parentid"
            fi
            # Sends password reminder email.
            export expiry FTPURL password REPLYTO="$SYSADMIN" restriction SYSADMIN username
            cat "$MAIL_PRM_PASSREMIND" | $ENVSUBST | $MAIL -a "$MAIL_PRM_PASSREMIND_MIME" -s "$MAIL_PRM_PASSREMIND_SUBJ" "$email"; RESULT_CODE=$?
            log "$MSG_PRM_MODIFYACC_PASSREMIND $username $email $RESULT_CODE"
            return $RESULT_CODE
        # Checks given password, must be match to continue.
        else
            local passhash="{md5}$(echo -n $password | $OPENSSL dgst -binary -md5 | $OPENSSL enc -base64)"
            if [ ! "$(getdetail 'passwd')" = "$passhash" ]; then 
                log "$MSG_PRM_MODIFYACC_AUTHFAIL $username"; return $EXIT_ERR 
            fi
        fi
    fi
    # Admin skips password checking, value discarded.
    password=""

    # Real modifications.
    local modified=0
    # New email is optional but formally checked.
    local email="$1"; shift
    if [ -z "$email" ]; then email="$(getdetail 'email')";
    else
        if [ ! -z "`echo "$email" | $SED "s/$MASKEMAIL//"`" ]; then 
            log "$MSG_PRM_MODIFYACC_WRONGEMAIL $email"
            return $EXIT_ERR 
        fi
        modified=1
    fi
    # New password is optional but formally checked.
    local password="$1"; shift
    local passclear passhash
    if [ -z "$password" ]; then 
        passclear="$(getdetail 'clear')";
        passhash="$(getdetail 'passwd')";
        password="$(echo -e "$passclear" | $OPENSSL enc -d -base64 | $OPENSSL enc -d -base64)"
    else
        if [ ! -z "`echo "$password" | $SED "s/$MASKPASS//"`" ]; then 
            log "$MSG_PRM_MODIFYACC_WRONGPASS $password" 
            return $EXIT_ERR
        fi
        passclear="$(echo -n $password | $OPENSSL enc -base64 | $OPENSSL enc -base64)"
        passhash="{md5}$(echo -n $password | $OPENSSL dgst -binary -md5 | $OPENSSL enc -base64)"
        modified=1
    fi
    # Expiry is optional but validated.
    local expiry="$1"; shift
    if [ -z "$expiry" ]; then expiry="$(getdetail 'expiry')";
    else
        # Changing expiry allowed only for administrators.
        isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_MODIFYACC_EXPIRY_NOADMIN"; return $EXIT_ERR; fi
        # Expiry validation via system 'date' command.
        if [ -z "$($DATE -d "$expiry" 2>/dev/null)" ]; then log "$MSG_PRM_MODIFYACC_WRONGEXPIRY $expiry"; return $EXIT_ERR; fi
        modified=1
    fi
    expiry="$($DATE '+%Y-%m-%d %H:%M:%S' -d "$expiry" 2>/dev/null)"
    # All details are processed.

    # Saves changes (if any) into database.
    if [ $modified -eq 1 ]; then 
        local MYSQL_UPDATEUSER="UPDATE users SET email='$email', passwd='$passhash', clear='$passclear', expiry='$expiry' WHERE userid='$username';"
        log "$username" "$(echo "$MYSQL_UPDATEUSER" | $MYSQL 2>&1; RESULT_CODE=$?)" "$RESULT_CODE"
        # Mail notification on sucess (will send credentials in plain text!).
        if [ $RESULT_CODE -eq 0 ]; then
            if [ -z "$expiry" -o "$expiry" = "NULL" -o "$expiry" = "0000-00-00 00:00:00" ]; then expiry="$MSG_PRM_MODIFYACC_NOEXPIRY"; fi
            local parentid="$(getdetail 'parentid')"
            local restriction
            if [ -z "$parentid" -o "$parentid" = "NULL" ]; then
                restriction="$MSG_PRM_MODIFYACC_NORESTRICT"
            else
                restriction="$MSG_PRM_MODIFYACC_RESTRICT $parentid"
            fi
            # Sends password reminder email.
            export expiry FTPURL password REPLYTO="$SYSADMIN" restriction SYSADMIN username
            cat "$MAIL_PRM_MODIFYACC" | $ENVSUBST | $MAIL -a "$MAIL_PRM_MODIFYACC_MIME" -s "$MAIL_PRM_MODIFYACC_SUBJ" "$email"; RESULT_CODE=$?
            log "$MSG_PRM_MAILSENT $email $RESULT_CODE"
        fi
    fi

    return $RESULT_CODE
}

# proftpm_removeaccount [-d] username [reason]
# Removes user from database.
# With -d (deletefiles) switch purges all files within his/her home
# Notifies deleted user by email (including reason if any).
function proftpm_removeaccount {

    # Permissions check: admin and mysql are required.
    isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi

    local RESULT_CODE=0

    # Delete files switch (-d) is optional.
    local option=`echo $1 | sed -n "s/^-\+//p"`; if [ ! -z "$option" ]; then shift; fi
    local deletefiles=0; if [ "$option" = "d" ]; then deletefiles=1; fi

    # Existing username is mandatory.
    local username="$1"; shift
    if [ -z "$username" ]; then log "$MSG_PRM_REMOVEACC_MISSINGUSER"; return $EXIT_ERR; fi
    # Loads user details into a simulated hash.
    # Some IFS tricks are used to avoid problems from fields containing spaces.
    local MYSQL_LISTACCOUNT="SELECT * FROM users WHERE userid='$username';"
    local ifs=$IFS; IFS=$MYSQL_IFS; local details=($(IFS=$ifs; echo "$MYSQL_LISTACCOUNT" | $MYSQL 2>/dev/null)); IFS=$ifs
    if [ -z "${details[0]}" ]; then log "$MSG_PRM_REMOVEACC_NOSUCHUSER $username"; return $EXIT_ERR; fi
    # Refuses deleting when user has pending tickets.
    if [ ! -z "$(proftpm_listtickets $username)" ]; then log "$MSG_PRM_REMOVEACC_PENDINGTICKETS $username"; return $EXIT_ERR; fi

    # Reason parameter is optional.
    local reason="$1"; shift

    # Deleting user 1st from database.
    local MYSQL_DELACCOUNT="DELETE FROM users WHERE userid='$username';"
    log "$username" "$(echo "$MYSQL_DELACCOUNT" | $MYSQL 2>&1; RESULT_CODE=$?)" "$RESULT_CODE"
    if [ $RESULT_CODE -eq 0 ]; then
        # On success and with deletefiles flag removes files also (if any)
        local homedir="$(getdetail 'homedir')"
        if [ $deletefiles -eq 1 ]; then
            # Empty, not within $FTPROOT or unwritable homedir is suspicious and logged.
            if [ -z "$homedir" -o $((( ${#homedir}-${#FTPROOT}-1 ))) -le 0 -o $(expr match "$homedir" "$FTPROOT") -eq 0  -o ! -w "$homedir" ]; then
                log "$MSG_PRM_REMOVEACC_HOMEPROBLEM" "$homedir"
            else
                # Actually deleting.
                rm -R "$homedir" 2>&1 1>/dev/null; RESULT_CODE=$?
                if [ $RESULT_CODE -ne $EXIT_OK ]; then log "$MSG_PRM_REMOVEACC_HOMEPROBLEM" "$homedir" "$RESULT_CODE"
                else log "$MSG_PRM_REMOVEACC_HOMEDELETED" "$homedir" "$RESULT_CODE"; fi
            fi
        else
            # Logs keeping home directory.
            log "$MSG_PRM_REMOVEACC_HOMEKEEPING" "$homedir"
        fi
        RESULT_CODE=0
        # Notifies user by email.
        local email="$(getdetail 'email')"
        if [ -z "$reason" ]; then reason="$MSG_PRM_REMOVEACC_DEFAULTREASON"; fi
        export FTPURL password reason REPLYTO="$SYSADMIN" SYSADMIN username
        cat "$MAIL_PRM_REMOVEDACC" | $ENVSUBST | $MAIL -a "$MAIL_PRM_REMOVEDACC_MIME" -s "$MAIL_PRM_REMOVEDACC_SUBJ" "$email"; RESULT_CODE=$?
        log "$MSG_PRM_MAILSENT $email $RESULT_CODE"
    fi

    return $RESULT_CODE
}

# proftpm_removetickets -rt | --removetickets [ownername]
# Wrapper function to remove tickets.
# With ownername given removes all tickets according to ownername.
# Without ownername removes all expired tickets.
function proftpm_removetickets {

    # Permissions check: admin and mysql are required.
    isadmin; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi
    ismysql; if [ $? -eq $EXIT_ERR ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi

    local RESULT_CODE=0

    # Ownername is optional.
    local ownername="$1"; shift
    if [ ! -z "$ownername" ]; then
        # With ownername enumerates all tickets according to ownername.
        local ticket
        local MYSQL_LIST_OWNED="SELECT userid FROM users WHERE parentid='$ownername' ORDER BY userid;"
        echo "$MYSQL_LIST_OWNED" | $MYSQL 2>&1 | $SED "s/^userid$//"  | while read ticket
        do
            # Deletes tickets via proftpm_removeaccount leaving files.
            if [ ! -z "$ticket" ]; then
                proftpm_removeaccount "$ticket"
            fi
        done
    else
        # Without ownername enumerates all expired tickets.
        local ticket
        local MYSQL_LIST_EXPIRED="SELECT userid FROM users WHERE parentid IS NOT NULL AND NOT parentid='' AND expiry < NOW();"
        echo "$MYSQL_LIST_EXPIRED" | $MYSQL 2>&1 | $SED "s/^userid$//"  | while read ticket
        do
            # Deletes enumerated tickets via proftpm_removeaccount
            # leaving files and sending a reason about expiry.
            if [ ! -z "$ticket" ]; then
                proftpm_removeaccount "$ticket" "$MSG_PRM_REMOVETICKETS_REASON"
            fi
        done
        RESULT_CODE=0
    fi

    return $RESULT_CODE
}

# Helper function to set up service
function proftpm_configure {

    # Must be root.
    if [ ! $UID -eq 0 ]; then log "$MSG_PRM_UNABLEFN"; return $EXIT_ERR; fi

    local datetime=$($DATETIME)

    # Logs handling: creation and rotation.
    # Must be the 1st step to enable logging configuration process itself.
    if [ ! -f "$LOGFILE" ]; then
        touch $LOGFILE; chown $FTPUSER:$FTPGROUP $LOGFILE; chmod 660 $LOGFILE
        log "$MSG_PRM_CONF_LOGFILE_CREATED" "$LOGFILE"
    else
        log "$MSG_PRM_CONF_LOGFILE_EXISTS" "$LOGFILE"
    fi
    if [ ! -f "$ROTATEFILE" ]; then
        export datetime FTPGROUP FTPUSER LOGFILE
        cat "$TEMPLATEROTATE" | $ENVSUBST > "$ROTATEFILE"; chmod 644 "$ROTATEFILE"
        log "$MSG_PRM_CONF_ROTATEFILE_CREATED" "$ROTATEFILE"
    else
        log "$MSG_PRM_CONF_ROTATEFILE_EXISTS" "$ROTATEFILE"
    fi

    # Stops daemon if running.
    log $($DAEMON stop 2>&1)

    # ProFTPm customisable defaults by template.
    if [ ! -r "$DEFAULTS" ]; then
        cp -p "$TEMPLATEDEFAULTS" "$DEFAULTS"
        log "$MSG_PRM_CONF_DEFAULTS_CREATED" "$DEFAULTS"
    else
        log "$MSG_PRM_CONF_DEFAULTS_EXISTS" "$DEFAULTS"
    fi
    # ProFTPm customisable mounts by template
    if [ ! -r "$MOUNTS" ]; then
        cp -p "$TEMPLATEMOUNTS" "$MOUNTS"
        log "$MSG_PRM_CONF_MOUNTS_CREATED" "$MOUNTS"
    else
        log "$MSG_PRM_CONF_MOUNTS_EXISTS" "$MOUNTS"
    fi

    # ProFTPd configuration with MySQL authentication.
    local mysql_usercheck="select user from mysql.user where user='$MYSQL_USER';"
    # Getting MySQL credentials (if any).
    local mysqlpassword=""
    if [ -r "$MYSQL_DEFAULTSFILE" ]; then
        mysqlpassword="$(cat "$MYSQL_DEFAULTSFILE" | $GREP 'password' | $SED 's/.*= //')"
    fi
    if [ -z "$mysqlpassword" ]; then mysqlpassword=$($MAKEPASSWD); fi
    # MySQL user creation - skip if user exists.
    if [ ! -z "$(echo "$mysql_usercheck" | $MYSQL_ROOT 2>&1)" ]; then
        log "$MSG_PRM_CONF_MYSQLUSER_EXISTS $MYSQL_USER"
    else
        # Generating dedicated MySQL user and a specific client config file.
	local mysql_usercreate="grant all privileges on $MYSQL_DATABASE.* to '$MYSQL_USER'@'localhost' identified by '$mysqlpassword';"
        echo "$mysql_usercreate" | $MYSQL_ROOT
        # Saving credentials into specific file readable by $MYSQL_LINUXUSERS only.
        export datetime mysqlpassword MYSQL_DATABASE MYSQL_USER
        rm $MYSQL_DEFAULTSFILE >/dev/null 2>&1
        cat "$TEMPLATESQLCNF" | $ENVSUBST > "$MYSQL_DEFAULTSFILE"
        # Credentials: root:root 600 and setfacl -m:user:r.
        chown root:root "$MYSQL_DEFAULTSFILE"; chmod 600 "$MYSQL_DEFAULTSFILE"
        for linuxuser in $MYSQL_LINUXUSERS
        do
            $SETFACL -m u:$linuxuser:r "$MYSQL_DEFAULTSFILE"
        done
        log "$MSG_PRM_CONF_MYSQLUSER_CREATED $MYSQL_DEFAULTSFILE"
    fi
    # Database setup/upgrade SQL script isn't destructive, we run always.
    log "$MSG_PRM_CONF_MYSQLSCRIPT" "$(cat "$TEMPLATESQL" | $MYSQL_ROOT 2>&1)"
    # ProFTPd configuration from template to use MySQL backend.
    if [ ! -f "$CONFIGFILE" ]; then
        export datetime CONFIGSQLFILE FTPGROUP FTPGID FTPROOT FTPUID FTPUSER LOGDIR MODULESFILE mysqlpassword SKELETON
        cat "$TEMPLATECONF" | $ENVSUBST > "$CONFIGFILE"; chmod 640 "$CONFIGFILE"
        log "$MSG_PRM_CONF_CONFFILE_CREATED" "$CONFIGFILE"
    else
        log "$MSG_PRM_CONF_CONFFILE_EXISTS" "$CONFIGFILE"
    fi
    # ProFTPd MySQL authentication setup from separated template.
    if [ ! -f "$CONFIGSQLFILE" ]; then
        export datetime MYSQL_DATABASE MYSQL_USER mysqlpassword
        cat "$TEMPLATECONFSQL" | $ENVSUBST > "$CONFIGSQLFILE"; chmod 640 "$CONFIGSQLFILE"
        log "$MSG_PRM_CONF_CONFSQLFILE_CREATED" "$CONFIGSQLFILE"
    else
        log "$MSG_PRM_CONF_CONFSQLFILE_EXISTS" "$CONFIGSQLFILE"
    fi

    # ProFTPd modules setup from template.
    if [ ! -f "$MODULESFILE" ]; then
        export datetime
        cat "$TEMPLATEMOD" | $ENVSUBST > "$MODULESFILE"; chmod 640 "$MODULESFILE"
        log "$MSG_PRM_CONF_MODFILE_CREATED" "$MODULESFILE"
    else
        log "$MSG_PRM_CONF_MODFILE_EXISTS" "$MODULESFILE"
    fi

    # FTP storage configuration.
    # Creating storage if absent.
    if [ ! -d "$FTPROOT" ]; then
        log $(mkdir -p -v "$FTPROOT"; chown $FTPUSER:$FTPGROUP "$FTPROOT"; chmod 2750 "$FTPROOT")
        # Disabling other's access by default.
        $SETFACL -d -m o::- "$FTPROOT"
    fi

    # Web access configuration.
    # Apache-2 configuration file from template.
    if [ ! -f "$APACHECONFIGDIR/$WEBURL" ]; then
        export APACHELOGDIR APACHEUSERDIR APACHEWEBIP APACHEWEBPORT
        export datetime FTPROOT FTPURL FTPPUBLICDIR PHPLOGDIR PHPLOGS
        export SYSADMIN WEBCUST WEBROOT WEBURL
        cat "$TEMPLATEHTTP" | $ENVSUBST > "$APACHECONFIGDIR/$WEBURL"
        log "$MSG_PRM_CONF_HTTPFILE_CREATED" "$APACHECONFIGDIR/$WEBURL"
    else
        log "$MSG_PRM_CONF_HTTPFILE_EXISTS" "$APACHECONFIGDIR/$WEBURL"
    fi
    # Customisable web folders symlinked to template if are absent.
    if [ ! -d "$WEBCUST" ]; then log $(mkdir -v -m 755 "$WEBCUST"); fi
    ls -1 "$WEBTEMPLATE" | while read webfolder
    do
        webfolder=$($BASENAME "$webfolder")
        if [ ! -e "$WEBCUST/$webfolder" ]; then
            log $(ln -v -s "$WEBTEMPLATE/$webfolder" "$WEBCUST/$webfolder")
        fi
    done

    # Finally start ProFTPd again.
    log $($DAEMON start)
    return
}

# Helper function: gets a simulated hash value from 'details' array.
# 'st all kexs then all values stored in array.
function getdetail {

    local key="$1"; shift
    local i pairs

    let pairs=${#details[*]}/2
    for i in $(seq $pairs)
    do
        if [ "${details[i-1]}" = "$key" ]; then
            echo "${details[i-1+$pairs]}"; return $EXIT_SUCC
        fi
    done
    return $EXIT_ERR
}

# Helper function: checks an item within a list
function inlist {

    local value="$1"; shift
    local list="$1"; shift
    local item

    if [ -z "$value" -o -z "$list" ]; then return $EXIT_ERR; fi
    for item in $list
    do 
        if [ "$item" = "$value" ]; then return $EXIT_OK; fi
    done
    return $EXIT_ERR
} 

# Helper function: current Linux user is admin user?
function isadmin {

    local username=$(whoami)

    # root is OK
    if [ ! $UID -eq 0 ]; then
        # admin users list is OK
        if [ `inlist "$username" "$ADMINUSERS"; echo $?` -eq $EXIT_ERR ]; then
            # admin group list is OK
            if [ `inlist "$ADMINGROUP" "$(groups "$username" | sed "s/^\(.*: \)\(.*\)$/\2/")"; echo $?` -eq $EXIT_ERR ]; then
                return $EXIT_ERR
            fi
        fi
    fi
    return $EXIT_OK
}

# Helper function: current Linux user accesses the MySQL?
function ismysql {

    # root is OK
    if [ ! $UID -eq 0 ]; then
        # issues a test query
        if [ -z "`echo -e 'SHOW DATABASES' | $MYSQL 2>/dev/null`" ]; then return $EXIT_ERR; fi
    fi

    return $EXIT_OK
}

# Helper function: logs with timestamp into a logfile and native to STD_ERR.
function log {

    if [ -f "$LOGFILE" -a -w "$LOGFILE" ]; then
        echo -e "[$($DATETIME)] $(whoami) ${FUNCNAME[1]} $*" >>$LOGFILE
        echo -e "$*" >&2
    else
        echo -e "[$($DATETIME)] $(whoami) ${FUNCNAME[1]} $*" >&2
    fi
    return
}

# Helper function: generates a simple pseudorandom alphanumeric string
# using a slice from /dev/urandom and md5 hashing.
# You may specify length within 1-31 characters.
function pseudorandom {

    local randombytes randomstring
    local length=${1:-10}

    cat /dev/urandom | while read -s -n$length randombytes
    do
        randomstring="$(echo -e "$randombytes" | $OPENSSL dgst -md5)"
        echo "${randomstring:1:length}"
        break
    done
    return
}


#
# Main program
#

# Handling languages
option=`echo $1 | $SED -n "s/^-\+//p"` 
if [ "$option" == "l" -o "$option" == "language" ]; then shift; 
    language="$1"; shift
    # Includes language file if exists - overdefines messages
    if [ -r "$LANGUAGEDIR/$language/messages" ]; then . "$LANGUAGEDIR/$language/messages"; fi
fi

# Handling options
option=`echo $1 | $SED -n "s/^-\+//p"`; if [ ! -z "$option" ]; then shift; fi
case "$option" in
    "ca"|"createaccount")
        proftpm_createaccount "$@"; exit $?
    ;;
    "ct"|"createticket")
        proftpm_createticket "$@"; exit $?
    ;;
    "la"|"listaccounts")
        proftpm_listaccounts "$@"; exit $?
    ;;
    "lt"|"listtickets")
        proftpm_listtickets "$@"; exit $?
    ;;
    "ma"|"modifyaccount")
        proftpm_modifyaccount "$@"; exit $?
    ;;
    "ra"|"removeaccount")
        proftpm_removeaccount "$@"; exit $?
    ;;
    "rt"|"removetickets")
        proftpm_removetickets "$@"; exit $?
    ;;
    "v"|"version")
	echo -e "$MSG_PRM_VERSION"; exit $?
    ;;
    "configure")
        proftpm_configure "$@"; exit $?
    ;;
    *)
        echo -e "$USAGE"; exit $EXIT_ERR
    ;;
esac
